package MegaSquirt::VE;
################################################################################
=head1 NAME

MegaSquirt::VE - The MegaSquirt class for dealing with VE tables

=head1 VERSION

Release Version: 0.01

=cut

our $VERSION = "0.01";

=head1 DATE

$Date: 2004/01/22 21:04:55 $

=head1 SYNOPSIS

  use MegaSquirt::VE;

=head1 DESCRIPTION

The MegaSquirt::VE Class is a class for dealing with VE Tables

=cut

################################################################################
use strict;

use MegaSquirt::Exception;
use Data::Dumper;

use base qw(MegaSquirt);

use constant MEM_FIELDS => qw(
			      EVEME
			      UserRev
			      UserComment
			      Date
			      Time
			      Page
			      rpm_range
			      kpa_range

			      filename
			      _matrix
			      _rpms
			      _kpas
			      max_rpm_index
			      max_kpa_index
			      );

use constant FIELDS => (MEM_FIELDS);
use constant ID_FIELD => 've_id';

my $MAX_INDEX = 7;
our $DEBUG = 0;
my $gen = 'MegaSquirt::Exception::General';

use fields (FIELDS);


my $VALIDATION = {
    ve_bin => { required => 1 },
};

sub validate {
    my __PACKAGE__ $self = fields::new(shift);
    my ($field,$val) = @_;
    
    if ($VALIDATION->{$field}->{required}) {
	return 'Field required' unless $val;
    } else {
	return undef unless $val;
    }

    if (defined $VALIDATION->{$field}->{method}) {
	my $rv;
	foreach my $method (@{$VALIDATION->{$field}->{method}}) {
	    $rv = $self->$method($val);
	    return $rv if $rv;
	}
	return $rv;
    }
    return undef;
}

foreach my $field (ID_FIELD, FIELDS) {
    no strict 'refs';
    *$field = sub {
	my __PACKAGE__ $self = shift;
	$self->_set({ $field => shift }) if @_;
	return $self->{$field};
    }
}

sub new {
    my __PACKAGE__ $self = fields::new(shift);
    $self->SUPER::new();
    $self->SUPER::_init();
    $self->_init(shift);
    return $self;
}


sub nearest_bin {
    my __PACKAGE__ $self = shift;
    my $rpm = shift;
    my $kpa = shift;

    my $rpm_bin = $self->nearest($rpm,$self->_rpms);
    my $kpa_bin = $self->nearest($kpa,$self->_kpas);
    return ($rpm_bin, $kpa_bin);
}

sub nearest {
    my __PACKAGE__ $self = shift;
    my $value = shift;
    my $array = shift;

    my $i = 0;
    my ($low,$high);
    return $array->[0] if $value < $array->[0];
    foreach (@$array) { 
	if ($_ < $value) {
	    $i++;
	    next;
	}
	$low = $array->[$i - 1] || $array->[0];
	$high = $array->[$i] || $array->[$i - 1];
#	return $array->[$i] || $array->[$i - 1];
    }
 
    my $rv;
    eval {
	my $low_delta = $value - $low;
	my $high_delta = $high - $value;
	if ($low_delta > $high_delta) {
	    $rv = $high;
	} elsif ($high_delta > $low_delta) {
	    $rv = $low 
	} else {
	    $rv = $low;
	}
    };
    return $array->[0] if $@;
    return $rv;

    return $array->[-1];
}


sub new_from_file {
    my __PACKAGE__ $self = fields::new(shift);
    my $args = shift;
    $self->SUPER::new();
    $self->SUPER::_init();
    $self->_init($args);
    
    my $filename = $args->{filename} || 
	throw $gen({ error => 'Filename Required' });

    #open file and slurp VE into internal data structure
    eval {
	$self->_read_file({ filename => $filename });
    };

    warn $@ if $@;
    throw $gen({ error => "File parsing error, $@" }) if $@;

    $self->_is_dirty('0');
    
    return $self;
}

sub _init {
    my __PACKAGE__ $self = shift;
    my $args = shift;
    $self->_set($args) if ref $args eq 'HASH';
}

sub _read_file {
    # this routine will break if the VEX format changes at all
    my __PACKAGE__ $self = shift;
    my $args = shift;
    my $filename = $args->{filename};

    open(VEX_FILE,$filename) || die 'unable to open file ' . $filename;

    my @lines;
    
    while (<VEX_FILE>) {
	push(@lines,$_);
    }
    close VEX_FILE;

    #GET META
    eval {
	$self->EVEME($1) if $lines[0] =~ m/^EVEME\s*(\d*\.\d*)/;
    };
    eval {
	$self->UserRev($1) if $lines[1] =~ m/^UserRev\:\s*(\d*\.\d*)/;
    };
    eval {
	$self->UserComment($1) if $lines[2] =~ m/^UserComment\:\s*(.*)/;
    };
    eval {
	$self->Date($1) if $lines[3] =~ m/^Date\:\s*(\d{2}\-\d{2}\-\d{4})/;
    };
    eval {
	$self->Time($1) if $lines[4] =~ m/^Time\:\s*(.*)/;
    };
    eval {
	$self->Page($1) if $lines[5] =~ m/^Page\s(.*)/;
    };

    my $i = 6;
    #skip any extra lines
    while ($lines[$i] !~ /RPM\sRange/i) {
	$i++;
    }

    eval {
	$self->rpm_range($1) if $lines[$i] =~ m/\[\s?(\d*)\]/;
    };
    
    # somewhat dynamic to handle n x n tables. Not fully implemented.
    $i = 7;
 
    my @rpm;
    my $j = $i;
    for ($j .. ($j + $self->rpm_range -1)) {
	$lines[$_] =~ m/\=\s*(\d*)\s/;
	push(@rpm,$1);
	$i++;
    }

    eval {
	$self->kpa_range($1) if $lines[$i] =~ m/\[\s?(\d*)\]/;
    };

    $i++;
    my @kpa;
    my $k = $i;
    for ($k .. ($k + $self->kpa_range -1)) {
	$lines[$_] =~ m/\=\s*(\d*)\s/;
	push(@kpa,$1);
	$i++;
    }

    $i++; #skip VE TABLE ROW
    $i++; #skip header row
    my @rows;
    for ($i .. $i + $self->rpm_range) {
	$lines[$_] =~ m/\=\s*(\d.*\d)\s/;	
	my $row = $1;
	my @vals = split(/\s\s*/,$row);
	push(@rows,\@vals);
    }

    # this is just for ease of use
    # flip the matrix for rpm=x,kpa=y ($m->[rpm]->[kpa])

    $self->max_rpm_index($#rpm);
    $self->max_kpa_index($#kpa);
    
    my $struc = $self->_flip_matrix(\@rows);;
    eval {
	$self->load_matrix({ rpms => \@rpm,
			     kpas => \@kpa,
			     rows => $struc,
			 });
    };
    warn $@ if $@;
    
}

sub space_pad {
    # this routine helps accomodate fixed length format
    my $val = shift;
    my $count = shift || 4; #default
    my @val_parts = reverse(split('',$val));
    my @blocks;
    for (0 .. $count -1) {
	my $part = defined($val_parts[$_]) ? $val_parts[$_] : ' ';
	unshift(@blocks, $part);
    }
    return join('',@blocks);
}
    

sub out {
    # build in versioning?
    my __PACKAGE__ $self = shift;
    my $string;

    my $padded_rpm = pad($self->rpm_range,1);
    my $padded_kpa = pad($self->kpa_range,1);
    $string .= <<END;
EVEME $self->{EVEME}
UserRev: $self->{UserRev}
UserComment: $self->{UserComment}
Date: $self->{Date}
Time: $self->{Time}
Page $self->{Page}
VE Table RPM Range              [$padded_rpm]
END

    for (0 .. $self->{max_rpm_index}) {
	$string .= qq(   [  $_] =) . space_pad($self->_rpms->[$_],4) . "\n";
    }


    $string .= <<END;
VE Table Load Range (MAP)       [$padded_kpa]
END

    for (0 .. $self->{max_kpa_index}) {
	$string .= qq(   [  $_] =) . space_pad($self->_kpas->[$_],4) . "\n";
    }

    $padded_rpm = pad($self->rpm_range,2);
    $padded_kpa = pad($self->kpa_range,2);

    $string .= qq(VE Table                        [$padded_rpm][$padded_kpa]\n);

    my @rpm_labels;
    for (0 .. $self->max_rpm_index) {
	push(@rpm_labels, '[' . pad($_,2) . ']');
    }
    my $rpm_labels = join(' ',@rpm_labels);
    $string .= qq(           $rpm_labels\n);

    foreach my $kpa (0 .. $self->max_kpa_index) {
	my $table_line = q(   [) . pad($kpa,2) . q(] =);
	my @vals;
	foreach my $rpm (0 .. $self->max_rpm_index) {
	    my $pad = $rpm == 0 ? 4 : 5;
	    push(@vals, pad($self->_matrix->[$rpm]->[$kpa],$pad));
	}
	
	$string .= $table_line . join('',@vals) . "\n";
    }
    $string .= "\n";
    return $string;
}

sub _old_read_file {
    my __PACKAGE__ $self = shift;
    my $args = shift;
    my $filename = $args->{filename};

    open(VEX_FILE,$filename) || die 'unable to open file ' . $filename;

    my @lines;
    
    while (<VEX_FILE>) {
	push(@lines,$_);
    }
    close VEX_FILE;
    my @rpm;
    for (7 .. 14) {
	$lines[$_] =~ m/\=\s*(\d*)\s/;
	push(@rpm,$1);
    }
    my @kpa;
    for (16 .. 23) {
	$lines[$_] =~ m/\=\s*(\d*)\s/;
	push(@kpa,$1);
    }
    
    my @rows;
    for (26 .. 33) {
	$lines[$_] =~ m/\=\s*(\d.*\d)\s/;	
	my $row = $1;
	my @vals = split(/\s\s*/,$row);
	push(@rows,\@vals);
    }

    # this is just vanity
    # flip the matrix for rpm=x,kpa=y ($m->[rpm]->[kpa])
    my $struc = $self->_flip_matrix(\@rows);;
    eval {
	$self->load_matrix({ rpms => \@rpm,
			     kpas => \@kpa,
			     rows => $struc,
			 });
    };
    warn $@ if $@;
}

sub _flip_matrix {
    my $self = shift;
    my $struc = shift;

    my $new_struc;
    # flip the axis
    for my $val1 (0 .. $self->max_rpm_index) {
	for my $val2 (0 .. $self->max_kpa_index) {
	    next unless $struc->[$val2]->[$val1];
	    $new_struc->[$val1]->[$val2] = $struc->[$val2]->[$val1];
	}
    }
    return $new_struc;
}

sub pad {
    my $val = shift;
    my $pad = shift;
    for (length($val) .. $pad) { 
	 $val = ' ' . $val;
     }
    return $val;
}

sub dump_matrix {
    my __PACKAGE__ $self = shift;

    my $flipped = $self->_flip_matrix($self->_matrix);
    my @rpms;
    foreach (@{$self->_rpms}) {
	push(@rpms, pad($_,1));
    }
    
    my $out;
    $out .= '      ' . join('--',@rpms) . "\n";
    $out .= "     --------------------------------\n";
    my $i = 0;
    foreach my $row (@{$flipped}) {
	$out .= pad($self->_kpas->[$i],3) . "|";
	foreach my $val (@$row) {
	    $out .= "-$val-";
	}
	$out .= "\n";
	$i++;
    }
    $out .= "\n";

}


sub redistribution {
    my __PACKAGE__ $self = shift;   
    my @y = @{shift()};
    my $new_size = shift() - 1 
	|| die $gen->throw("Error: impossible size for redimension!");
    my @x = (0 .. $#y);
    
    require Math::Spline;
    my $spline=new Math::Spline(\@x,\@y);

    my $incr = $#y/$new_size;
    my @r;
    my $pos = 0;
    for (0 .. $new_size) {
	push(@r, int($spline->evaluate($pos)));
	$pos += $incr;
    }

    return \@r;
}




sub convert_map {
    my __PACKAGE__ $self = shift;   
    my $args = shift;
    # this will spline the original values and provide 
    # new data points
    my $new_size_r = $args->{rpm_size};
    my $new_size_k = $args->{kpa_size};
    my $rpms = $self->redistribution($self->_rpms,$new_size_r);
    my $kpas = $self->redistribution($self->_kpas,$new_size_k);

    my $new_ve = __PACKAGE__->new;
    $new_ve->_rpms($rpms);
    $new_ve->_kpas($kpas);

    my @new_matrix;
    foreach my $r (@$rpms) {
	my @nrpm;
	foreach my $k (@$kpas) {
	    my $calc_ve = $self->round($self->calculate_ve({ rpm => $r,
							     kpa => $k }));
	    push(@nrpm, $calc_ve);
	}
	push @new_matrix, \@nrpm;
    }
    $new_ve->_matrix(\@new_matrix);
    $new_ve->max_rpm_index($new_size_r - 1);
    $new_ve->max_kpa_index($new_size_k - 1);
    $new_ve->rpm_range($new_size_r);
    $new_ve->kpa_range($new_size_k);


    $new_ve->EVEME($self->EVEME);
    $new_ve->UserRev($self->UserRev);
    $new_ve->UserComment("converted to $new_size_r x $new_size_k");
    $new_ve->Date($self->Date);
    $new_ve->Time($self->Time);
    $new_ve->Page($self->Page);
    
    return $new_ve;
}


sub as_html {
    my __PACKAGE__ $self = shift;

    my $flipped = $self->_flip_matrix($self->_matrix);
    my @rpms;
    foreach (@{$self->_rpms}) {
	push(@rpms, $_);
    }
    
    my $out;
    $out .= qq(<table border="1" cellspacing="3" cellpadding="3">\n);
    $out .= qq(<tr><th>);

    $out .= '<tr><th></th><th>' . join('</th><th>',@rpms) . "</th></tr>\n";

    my $i = 0;
    foreach my $row (@{$flipped}) {
 
	$out .= qq(<tr><th>) . $self->_kpas->[$i] . qq(</th>);
	foreach my $val (@$row) {
	    $out .= "<td>$val</td>";
	}

	$out .= "</tr>\n";
	$i++;
    }
    $out .= "</table>";
}


sub load_matrix {
    my __PACKAGE__ $self = shift;
    my $args = shift;
    # verify that the rpm and kpa numbers make sense!
    eval { $self->legal_bin_labels($args->{rpms}); };
    eval { $self->legal_bin_labels($args->{kpas}); };
    throw $gen({ error => "$@" }) if $@;

    $self->_rpms($args->{rpms});
    $self->_kpas($args->{kpas});

    $self->_matrix($args->{rows}); 


}

sub legal_bin_labels {
    my __PACKAGE__ $self = shift;
    my $self = shift;
    my $arr = shift;
    my $i;
    foreach (@$arr) {
	throw $gen({ error => "$@" }) if $@;
	$i = $_;
    }
}


sub bin {
    my __PACKAGE__ $self = shift;
    my ($rpm, $kpa, $is_index, $set) = @_;

    my $rpm_index = $rpm;
    my $kpa_index = $kpa;
    unless ($is_index) {
	$rpm_index = _in($rpm,$self->_rpms);
	$kpa_index = _in($kpa,$self->_kpas);
    }
    return undef unless defined($rpm_index);
    return undef unless defined($kpa_index);
    if(defined($set)) {
	$self->_matrix->[$rpm_index]->[$kpa_index] = $set;
    }
    return $self->_matrix->[$rpm_index]->[$kpa_index];
}

sub _in {
    my $value = shift;
    my $array = shift;
    my $i = 0;
    for ($i .. ($array->[-1])) { return $i if $value eq $array->[$i]; $i++; }
    return undef;
}

sub rpm {
    my __PACKAGE__ $self = shift;
    my $rpm_index = shift;
    return $self->_rpms->[$rpm_index];
}

sub kpa {
    my __PACKAGE__ $self = shift;
    my $kpa_index = shift;
    return $self->_kpas->[$kpa_index];
}


sub calculate_ve {
# This method is designed to simulate the controller code.
# The number calculated here will be identical
# to the VE calculated by Megasquirt
    my __PACKAGE__ $self = shift;
    my $args = shift;
    my $DEBUG = 0;
    print "####################################################\n" if $DEBUG;
    my $rpm = $args->{rpm};
    my $kpa = $args->{kpa};

    # if the values are exact bin points, return the bin value
    return $self->bin($rpm,$kpa) if $self->bin($rpm,$kpa);
    
    # else interpolate what the VE should be
    my ($brpm,$bkpa);

    # find the bracketing bins
    ($brpm->{1},$brpm->{2}) = $self->bracketed($args->{rpm},$self->_rpms);
    ($bkpa->{1},$bkpa->{2}) = $self->bracketed($args->{kpa},$self->_kpas);

    #########################################################
    if ($DEBUG) {
	print qq(\n=-=-=-= Bracketing Bins =-=-=-=\n);
	print qq( rpm->{1} = ) . $self->rpm($brpm->{1}) . qq(, rpm->{2} = ) . $self->rpm($brpm->{2});
	print qq( kpa->{1} = ) . $self->kpa($bkpa->{1}) . qq(, kpa->{2} = ) . $self->kpa($bkpa->{2});
    }
    #########################################################


    #########################################################
    ## x1=rpm1, x2=rpm2, y1=VE(kpa1,rpm1), y2=VE(kpa1,rpm2)
    if ($DEBUG) {
    print qq(\n---- Interp 1 ----------------------\n);
    print qq(x1=rpm1, x2=rpm2, y1=VE(kpa1,rpm1), y2=VE(kpa1,rpm2)\n);
    print "x = $rpm\n";
    print "x1 = " . $self->rpm($brpm->{1}) . "\n";
    print "x2 = " . $self->rpm($brpm->{2}) . "\n";
    print "y1 = " . $self->bin($brpm->{1},$bkpa->{1},1) . "\n";
    print "y2 = " . $self->bin($brpm->{2},$bkpa->{1},1) . "\n";
    }
    #########################################################

    my $ve_low_kpa = $self->_interpolate({ 
	x => $rpm,
	x1 => $self->rpm($brpm->{1}),
	x2 => $self->rpm($brpm->{2}),
	y1 => $self->bin($brpm->{1},$bkpa->{1},1),
	y2 => $self->bin($brpm->{2},$bkpa->{1},1),
    });
    
    #########################################################
    ## x1=rpm1, x2=rpm2, y1=VE(kpa2,rpm1), y2=VE(kpa2,rpm2)
    if ($DEBUG) {
    print qq(\n---- Interp 2 -----------------\n);
    print qq(x1=rpm1, x2=rpm2, y1=VE(kpa2,rpm1), y2=VE(kpa2,rpm2)\n);
    print "x = $rpm\n";
    print "x1 = " . $self->rpm($brpm->{1}) . "\n";
    print "x2 = " . $self->rpm($brpm->{2}) . "\n";
    print "y1 = " . $self->bin($brpm->{1},$bkpa->{2},1) . "\n";
    print "y2 = " . $self->bin($brpm->{2},$bkpa->{2},1) . "\n";
    }
    #########################################################

    my $ve_high_kpa = $self->_interpolate({ 
	x => $rpm,
	x1 => $self->rpm($brpm->{1}),
	x2 => $self->rpm($brpm->{2}),
	y1 => $self->bin($brpm->{1},$bkpa->{2},1),
	y2 => $self->bin($brpm->{2},$bkpa->{2},1)
	});

    #########################################################
    ## x1=kpa1, x2=kpa2, y1=VE_FROM_STEP_4, y2=VE_FROM_STEP_5
    if ($DEBUG) {
    print qq(\n--- Interp 3 ----------------\n);
    print qq(x1=kpa1, x2=kpa2, y1=VE_FROM_STEP_4, y2=VE_FROM_STEP_5\n);
    print "x = $kpa\n";
    print "x1 = " . $self->kpa($bkpa->{1}) . "\n";
    print "x2 = " . $self->kpa($bkpa->{2}) . "\n";
    print "y1 = " . $ve_low_kpa . "\n";
    print "y2 = " . $ve_high_kpa . "\n";
    }
    #########################################################

    my $ve = $self->_interpolate({  x => $kpa,
				    x1 => $self->kpa($bkpa->{1}),	
				    x2 => $self->kpa($bkpa->{2}),
				    y1 => $ve_low_kpa,
				    y2 => $ve_high_kpa, });
    # round to int and return
    return $self->round($ve);
}

sub round {
    my __PACKAGE__ $self = shift;
    my $value = shift;
    my $rv;
    eval {
	my $partial = $value - int($value);
	$rv = $partial >= .5 ? int($value) + 1 : int($value);
    };
    return int($value) if $@;
    return $rv;
}



sub _interpolate {
    # linear
    my $self = shift;
    my $a = shift;
    my $Y;
    if ($DEBUG) {
    print <<END;
***************************************************************************
** Linear Interpolation - 2D
**
**            (y2 - y1)
**  Y = Y1 +  --------- * (x - x1)
**            (x2 - x1)
**
**   tmp1 = x1
**   tmp2 = x2
**   tmp3 = y1
**   tmp4 = y2
**   tmp5 = x
**   tmp6 = y
***************************************************************************

	my \$Y = $a->{y1} + 
	    ($a->{y2} - $a->{y1}) / ($a->{x2} - $a->{x1}) + ($a->{x} - $a->{x1});
    };
END
    }

    $Y;
    eval {
    # based on: Y = y1 + (y2 - y1)/(x2 - x1) * (x - x1)
	$Y = 
	    $a->{y1} + 
	    
	    (($a->{y2} - $a->{y1}) / ($a->{x2} - $a->{x1})) *
	    
	    ($a->{x} - $a->{x1});
    };
    # throw back a zero if illegal division
    $Y = 0 if $@ =~ /Illegal division by zero/;
    print "Y = $Y\n" if $DEBUG;
    return $Y;
}




sub _old_interpolate {
    # linear
    my $self = shift;
    my $a = shift;
    my $Y;
    if ($DEBUG) {
    print <<END;
***************************************************************************
** Linear Interpolation - 2D
**
**            (y2 - y1)
**  Y = Y1 +  --------- * (x - x1)
**            (x2 - x1)
**
**   tmp1 = x1
**   tmp2 = x2
**   tmp3 = y1
**   tmp4 = y2
**   tmp5 = x
**   tmp6 = y
***************************************************************************

	my \$Y = $a->{y1} + 
	    ($a->{y2} - $a->{y1}) / ($a->{x2} - $a->{x1}) + ($a->{x} - $a->{x1});
    };
END
    }

    $Y;
    eval {
    # based on: Y = y1 + (y2 - y1)/(x2 - x1) + (x - x1)
	$Y = 
	    $a->{y1} + 
	    
	    (($a->{y2} - $a->{y1}) / ($a->{x2} - $a->{x1})) +
	    
	    ($a->{x} - $a->{x1});
    };
    # throw back a zero if illegal division
    $Y = 0 if $@ =~ /Illegal division by zero/;
    print "Y = $Y\n" if $DEBUG;
    return $Y;
}


sub bracketed {
    my __PACKAGE__ $self = shift;
    my $DEBUG = 0;
    if ($DEBUG) {
	print "\n----------------------\n";
    }
    my $value = shift;
    my $array = shift;

    
    my $i = 0;
    my ($low, $high);

    return (0,1) if $value < $array->[0];
    return ($MAX_INDEX,$MAX_INDEX) if $value > $array->[$MAX_INDEX];
   
    
    for ($i = 0; $i <= $MAX_INDEX; $i++) {
	if ($DEBUG) {
	    print qq((i: $i, $array->[$i], value: $value)\n);
	}
	next unless (($value >= $array->[$i]) && ($value <= $array->[$i + 1]));
	print "$i next unless (($value > $array->[$i]) && ($value < $array->[$i + 1]))" if $DEBUG;

	$low = $i;
	$high = $i + 1;
	print "\nFOUND (low: $low, high: $high)\n" if $DEBUG;

	return ($low, $high);
    }
}

sub lookup {
    my __PACKAGE__ $self = fields::new(shift);
    my $args = shift;
    $self->SUPER::new();
    $self->SUPER::_init();
    $self->_init($args);

    $self->_is_dirty('0');
    return $self;
}


sub list {
    my ($class, $args) = @_;
    return _list($class, $args, undef);
}

sub list_ids {
    my ($class, $args) = @_;
    return _list($class, $args, 1);
}

sub delete {
    my __PACKAGE__ $self = shift;
}

sub save {
    my __PACKAGE__ $self = shift;
    if ($self->_get(ID_FIELD)) {
	$self->_update;
    } else {
	$self->_insert;
    }
}

sub _list {
    my ($class, $args, $ids_only) = @_;
}

sub _insert {
    my __PACKAGE__ $self = shift;
}

sub _update {
    my __PACKAGE__ $self = shift;
    return unless $self->_is_dirty;
}


sub voltage_to_af {
    my $self = shift;
    my $voltage = shift;
    return ((2* $voltage) + 10);
}

sub af_to_voltage {
    my $self = shift;
    my $af = shift;
    return 0 unless $af;
    return $af/2 -10;
}


=head1 AUTHOR

steve berry, E<lt>sberry@unideal.netE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2004 by steve berry

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 

=head1 SEE ALSO

NONE

=cut

1;
